import torch

from .fixgaugeqr import fixgauge_qr
from .dominanteigensolver import dominant_eigensolver


def left_orthonormalize(A, C, epsilon, maxiter=100):
    '''
    C*A = lambda*AL*C 
    '''
    chi, d = A.shape[0], A.shape[1]
    A = A.contiguous().view(chi, d * chi)

    def step(C):
        _, C = fixgauge_qr(C)
        C = C / C.norm()
        AC = (C @ A).view(chi * d, chi)
        AL, Cnew = fixgauge_qr(AC)
        lamb = Cnew.norm()
        Cnew = Cnew / lamb
        diff = torch.dist(Cnew, C)
        return AL.view(chi, d, chi), Cnew, lamb, diff

    AL, Cnew, lamb, diff = step(C)
    for it in range(maxiter):
        def Hopt(x):
            '''
            C = A^{T}_L C A
            '''
            C = x.view(chi, chi)
            AC = (C @ A).view(chi * d, chi)
            y = AL.view(chi * d, chi).t() @ AC
            return y.view(-1)

        _, C = dominant_eigensolver(Hopt, Cnew.contiguous().view(-1), diff / 10)
        C = C.view(chi, chi)

        AL, Cnew, lamb, diff = step(C)
        if (diff < epsilon):
            break
        else:
            C = Cnew
    return AL, C, lamb


def right_orthonormalize(A, C, epsilon):
    '''
    A*C = lambda*C*AR
    '''
    AL, C, lamb = left_orthonormalize(A.permute(2, 1, 0), C.permute(1, 0), epsilon)
    return AL.permute(2, 1, 0).contiguous(), C.permute(1, 0).contiguous(), lamb


if __name__ == '__main__':
    torch.manual_seed(42)
    chi = 30
    d = 2
    epsilon = 1E-12
    dtype = torch.float64
    A = torch.randn(chi, d, chi, dtype=dtype)

    AL, C, lamb = left_orthonormalize(A, torch.eye(chi, dtype=dtype), epsilon)
    print(
        torch.dist((C @ A.view(chi, d * chi)).view(chi, d, chi), lamb * (AL.view(chi * d, chi) @ C).view(chi, d, chi)))
    AL = AL.view(chi * d, chi)
    print(torch.dist(AL.t() @ AL, torch.eye(chi, dtype=dtype)))

    # AR, C, lamb = right_orthonormalize(A, torch.eye(chi,  dtype=dtype), epsilon)
    # print (torch.dist((A.view(chi*d, chi)@C).view(-1), lamb*(C@AR.view(chi, d*chi)).view(-1)))
    # AR = AR.view(chi, d*chi)
    # print (AR@AR.t())
